import re
from abc import abstractmethod
from collections import defaultdict
from functools import lru_cache
from typing import DefaultDict, Mapping, Union

from datasets import Dataset, load_dataset
from evaluate import load as load_metric
from llama_cpp import LlamaGrammar
from tqdm import tqdm

from models import Llama2, OpenAI
from utils import log_calls, logger

CLASSIFICATION_PROMPT = """
Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

### Instruction:
{instruction}.

### Input:
{input}

### Response:
"""


class Task:
    def __init__(
        self,
        model: Union[Llama2, OpenAI],
        validation_dataset: str,
        test_dataset: str,
        *,
        use_grammar: bool,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        self.model = model

        # whether we use the grammar to constrain the model output or not
        self.use_grammar = use_grammar

        self.validation_dataset = load_dataset(
            validation_dataset, split=validation_split
        )
        self.test_dataset = load_dataset(test_dataset, split=test_split)

    @abstractmethod
    def predict(self, prompt: str, *args, **kwargs):
        pass

    @abstractmethod
    def _evaluate(self, prompt: str, dataset):
        pass

    @log_calls("Evaluating validation dataset")
    def evaluate_validation(self, prompt: str):
        return self._evaluate(prompt, self.validation_dataset)

    @log_calls("Evaluating test dataset")
    def evaluate_test(self, prompt: str):
        return self._evaluate(prompt, self.test_dataset)

    @property
    @abstractmethod
    def metric_name(self):
        pass

    @property
    @abstractmethod
    def base_prompt(self):
        pass


# a simple grammar in GBNF notation which allows either positive or negative
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
SA_OUTPUTS = set(("positive", "negative"))


@lru_cache
def sa_grammar_fn(verbose: bool = False):
    return LlamaGrammar.from_string(
        "root ::= ({})".format("|".join(('"' + cons + '"') for cons in SA_OUTPUTS)),
        verbose=verbose,
    )


class SentimentAnalysis(Task):

    def __init__(
        self,
        model,
        validation_dataset: str,
        test_dataset: str,
        *,
        use_grammar: bool,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        super().__init__(
            model,
            validation_dataset,
            test_dataset,
            use_grammar=use_grammar,
            validation_split=validation_split,
            test_split=test_split,
        )

    def predict(self, prompt: str, text: str):
        # run model for inference using grammar to constrain output
        # TODO grammar also depends on prompt and vice-versa -> what are good labels?
        response = self.model(
            prompt=CLASSIFICATION_PROMPT.format(instruction=prompt, input=text),
            grammar=sa_grammar_fn() if self.use_grammar else None,
            chat=False if self.use_grammar else True,
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()
        # input(f"*** PREDICTION***\n\tText: {text}\n\tSentiment: {response}")

        return response

    def _evaluate(self, prompt: str, dataset: Dataset):
        sst2_labels = {"negative": 0, "positive": 1}

        results: DefaultDict[str, int] = defaultdict(int)
        dataset_iterator = tqdm(dataset, desc="evaluating prompt", leave=False)

        for datum in dataset_iterator:
            response = self.predict(prompt=prompt, text=datum["text"]).lower()

            if self.use_grammar:
                # model output is from label space
                answer_label = sst2_labels[response]
            else:
                answer_label = None
                for label in sst2_labels.keys():
                    if label in response:
                        answer_label = sst2_labels[label]
                        break
                else:
                    logger.warning(f"Invalid answer: {response}")
                    results["failed"] += 1
                    continue

            classification_result = (
                "incorrect" if answer_label != datum["label"] else "correct"
            )
            results[classification_result] += 1
            dataset_iterator.set_postfix(results)

        accuracy = results["correct"] / sum(results.values())
        return accuracy

    @property
    def metric_name(self):
        return "accuracy"

    @property
    def base_prompt(self):
        #  from the paper: RLPROMPT: Optimizing Discrete Text Prompts with Reinforcement Learning
        return """In this task, you are given sentences from movie reviews. The task is to classify a sentence as "’positive’" if the sentiment of the sentence is positive or as "’negative’" if the sentiment of the sentence is negative. Return label only without any other text."""


def grammar_continuous_with_arbitrary_end(sequence: list[str], quote: str = '"'):
    if len(sequence) == 0:
        return ""

    if quote is None:
        quote = ""

    escape_item = lambda text: text.replace('"', '\\"')

    grammar = quote + escape_item(sequence[0]) + quote
    for idx, item in enumerate(sequence[1:]):
        new_item = "(" + quote + escape_item(item) + quote + ")?"
        if idx == 0:
            grammar += new_item
        else:
            grammar = grammar[: -2 * idx] + new_item + grammar[-2 * idx :]
    return grammar


# a grammar in GBNF notation which takes the context into account
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
@lru_cache
def qa_grammar_fn(context: str, verbose: bool = False):
    grammar_str = "(\n"
    tokens = re.split(r"(\W+)", context.strip())[:-1]

    for i in range(len(tokens)):
        if i > 0:
            grammar_str += "\n|\n"
        grammar_str += "(" + grammar_continuous_with_arbitrary_end(tokens[i:]) + ")"
    grammar_str += "\n)"
    return LlamaGrammar.from_string(f"root ::= {grammar_str}", verbose=verbose)


class QuestionAnswering(Task):

    def __init__(
        self,
        model,
        validation_dataset: str,
        test_dataset: str,
        *,
        use_grammar: bool,
        validation_split: str = None,
        test_split: str = None,
    ) -> None:
        self.evaluation_prompt = """
        Below is an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.

        ### Instruction:
        {instruction}

        ### Context:
        {context}

        ### Question:
        {question}

        ### Response:
        """

        self.metric = load_metric("squad")

        super().__init__(
            model,
            validation_dataset,
            test_dataset,
            use_grammar=use_grammar,
            validation_split=validation_split,
            test_split=test_split,
        )

    def predict(self, prompt: str, context: str, question: str):
        # run model for inference

        grammar = None
        if self.use_grammar:
            # context-sensitive grammar
            try:
                grammar = qa_grammar_fn(context)
            except Exception as e:
                logger.exception(
                    "Could not create grammar for context (potentially maximum recursion depth exceeded), therefore we do not use a grammar for this sample.\nContext (with length %d): %s",
                    len(context),
                    context,
                    exc_info=e,
                )

        response = self.model(
            prompt=self.evaluation_prompt.format(
                instruction=prompt,
                context=context,
                question=question,
            ),
            grammar=grammar,
            chat=False if self.use_grammar else True,
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()
        # input(f"*** PREDICTION***\n\tContext: {context}\n\tQuestion: {question}\n\tAnswer: {response}")

        return response

    def _evaluate(self, prompt: str, dataset: Dataset):
        def replace_symbol_for_grammar(sample: Mapping):
            sample["context"] = sample["context"].replace("–", "-")
            sample["answers"]["text"] = [
                text.replace("–", "-") for text in sample["answers"]["text"]
            ]
            return sample

        if self.use_grammar:
            # NOTE: the LlamaGrammar has issues with symbol '–' therefore we replace all occurences with '-' (hyphen)
            dataset = dataset.map(replace_symbol_for_grammar, desc="Replacing symbols")

        dataset_iterator = tqdm(dataset, desc="evaluating prompt", leave=False)

        num_samples = 0
        f1 = 0.0
        em = 0
        for datum in dataset_iterator:
            answer = self.predict(
                prompt, context=datum["context"], question=datum["question"]
            ).lower()

            num_samples += 1
            result = self.metric.compute(
                predictions=[{"prediction_text": answer, "id": datum["id"]}],
                references=[{"answers": datum["answers"], "id": datum["id"]}],
            )
            f1 += result["f1"]
            em += result["exact_match"]

            dataset_iterator.set_postfix(
                {"f1": f1 / num_samples, "em": em / num_samples}
            )

        return f1 / num_samples

    @property
    def metric_name(self):
        return "f1"

    @property
    def base_prompt(self):
        # TODO find good prompt
        return """In this task, you are given contexts with questions. The task is to answer the question given the context. Return only the answer without any other text. Make sure that the answer is taken directly from the context."""
